# In order to compute masses for G by solving linear equations, we need to
# compute the spectral side for a certain list Lambda(G) of weights, some of
# which typically have rather large motivic weights.  For this reason, we don't
# want to compute the "old" part of the spectral side (inductively, so using
# masses for smaller groups) for all weights corresponding to motivic weight <=
# constant.  Rather, we will compute all weights for smaller groups which might
# occur in an Arthur-Langlands parameter with infinitesimal character in
# Lambda(G).

def subset_binary(A, m):
    res = []
    mask = 1
    for x in A:
        if mask&m:
            res.append(x)
        mask <<= 1
    return res

def subsets_motw_gt(L, min_motw):
    n = len(L)
    i = 0
    while i<n and L[i]>=min_motw:
        i += 1
    if i==0:
        return
    A = L[:i]
    B = L[i:]
    for mA in range(1, 1<<i):
        for mB in range(1<<(n-i)):
            yield subset_binary(A, mA)+subset_binary(B, mB)

def dw_mask(dw):
    mask = 0
    c = 1
    prev = 0
    for x in reversed(dw):
        if x==prev:
            mask |= c
        c <<= 1
        prev = x
    return mask

class pre_endo_ctx:

    def add_SO_odd(self, rank, dw):
        for i in range(len(self.SO_odd)+1, rank+1):
            self.SO_odd.append(set())
            self.SO_odd_bymask.append([[] for j in range(1<<i)])
        if dw in self.SO_odd[rank-1]:
            return
        self.SO_odd[rank-1].add(dw)
        m = dw_mask(dw)
        for j in range(1<<rank):
            if j&m != j:
                continue
            self.SO_odd_bymask[rank-1][j].append(dw)

    def add_Sp(self, rank, dw):
        for i in range(len(self.Sp)+1, rank+1):
            self.Sp.append(set())
            self.Sp_bymask.append([[] for j in range(1<<i)])
        if dw in self.Sp[rank-1]:
            return
        self.Sp[rank-1].add(dw)
        m = dw_mask(dw)
        for j in range(1<<rank):
            if j&m != j:
                continue
            self.Sp_bymask[rank-1][j].append(dw)

    def add_SO_even(self, rank, dw):
        hr = ZZ(rank/2)
        for i in range(len(self.SO_even)+1, hr+1):
            self.SO_even.append(set())
            self.SO_even_bymask.append([[] for j in range(1<<(2*i))])
        if dw in self.SO_even[hr-1]:
            return
        self.SO_even[hr-1].add(dw)
        m = dw_mask(dw)
        for j in range(1<<rank):
            if j&m != j:
                continue
            self.SO_even_bymask[hr-1][j].append(dw)

    # hw is a decreasing list of weights, e.g. [11] for Ramanujan Delta
    # enumerate all weights of cuspidal representations that might occur in an
    # endoscopic parameter with weight hw
    def endo_wts_SO_odd(self, hw):
        n = len(hw)
        for sub_hw in subsets_motw_gt(hw, self.min_motw):
            k = len(sub_hw)
            # sympl pi
            sub_dw = tuple(ZZ((sub_hw[i]+1)/2)-k+i for i in range(k))
            #print("SO_odd "+str(sub_dw)+" "+str(tuple(2*(sub_dw[i]+k-i)-1 for i in range(k))))
            self.add_SO_odd(k, sub_dw)
            r = int(k/2)
            # check if sub_hw can come from some pi[2]
            if not(all(sub_hw[2*i+1]==sub_hw[2*i]-2 for i in range(r))):
                continue
            if k%2:
                # pi[2] with pi odd orth
                if sub_hw[k-1]!=1:
                    continue
                sub_dw = tuple(ZZ((sub_hw[2*i]-1)/2)-r+i for i in range(r))
                #print("Sp "+str(sub_dw)+" "+str(tuple(2*(sub_dw[i]+r-i) for i in range(r))))
                self.add_Sp(r, sub_dw)
            elif not((k/2)%2):
                # pi[2] with pi even orth
                sub_dw = tuple(ZZ((sub_hw[2*i]-1)/2)-r+i+1 for i in range(r))
                #print("SO_even "+str(sub_dw)+" "+str(tuple(2*(sub_dw[i]+r-i-1) for i in range(r))))
                self.add_SO_even(r, sub_dw)

    def endo_wts_Sp(self, hw):
        n = len(hw)
        for sub_hw in subsets_motw_gt(hw, self.min_motw):
            k = len(sub_hw)
            # odd orth pi
            sub_dw = tuple(ZZ(sub_hw[i]/2)-k+i for i in range(k))
            self.add_Sp(k, sub_dw)
            if k%2:
                continue
            # even orth pi
            sub_dw = tuple(ZZ(sub_hw[i]/2)-k+i+1 for i in range(k))
            self.add_SO_even(k, sub_dw)
            # pi[2] with pi sympl
            r = ZZ(k/2)
            if not(all(sub_hw[2*i+1]==sub_hw[2*i]-2 for i in range(r))):
                continue
            sub_dw = tuple(ZZ(sub_hw[2*i]/2)-r+i for i in range(r))
            self.add_SO_odd(r, sub_dw)

    def endo_wts_SO_even(self, hw):
        n = len(hw)
        for sub_hw in subsets_motw_gt(hw, self.min_motw):
            k = len(sub_hw)
            # odd orth
            if sub_hw[k-1]==0 and k>1:
                sub_dw = tuple(ZZ(sub_hw[i]/2)-k+i+1 for i in range(k-1))
                self.add_Sp(k-1, sub_dw)
            if k%2:
                continue
            # even orth
            sub_dw = tuple(ZZ(sub_hw[i]/2)-k+i+1 for i in range(k))
            self.add_SO_even(k, sub_dw)
            # pi[2] with pi sympl
            r = ZZ(k/2)
            if not(all(sub_hw[2*i+1]==sub_hw[2*i]-2 for i in range(r))):
                continue
            sub_dw = tuple(ZZ(sub_hw[2*i]/2)-r+i for i in range(r))
            self.add_SO_odd(r, sub_dw)

    # wts is a list of weights in the format used for the explicit formula
    # only weights having motivic weight >=min_motw are kept, the idea being
    # that in motivic weight <min_motw we will compute for all weights anyway.
    def __init__(self, wts, min_motw):
        self.Sp = []
        self.Sp_bymask = []
        self.SO_odd = []
        self.SO_odd_bymask = []
        self.SO_even = []
        self.SO_even_bymask = []
        self.min_motw = min_motw
        for wt in wts:
            d = dimw(wt)
            if Mod(d,2)==1:
                # odd dim => orth
                dw = wt2dw_Sp(wt)
                n = len(dw)
                self.endo_wts_Sp([2*(dw[i]+n-i) for i in range(n)])
            elif Mod(wt[-1],2)==1:
                # sympl
                dw = wt2dw_SO_odd(wt)
                n = len(dw)
                self.endo_wts_SO_odd([2*(dw[i]+n-i)-1 for i in range(n)])
            else:
                # even orth
                dw = wt2dw_SO_even(wt)
                n = ZZ(len(dw)/2)
                self.endo_wts_SO_even([2*(dw[i]+2*n-i-1) for i in range(2*n)])
        while len(self.Sp)>0 and len(self.Sp[-1])==0:
            self.Sp.pop()
            self.Sp_bymask.pop()
        while len(self.SO_odd)>0 and len(self.SO_odd[-1])==0:
            self.SO_odd.pop()
            self.SO_odd_bymask.pop()
        while len(self.SO_even)>0 and len(self.SO_even[-1])==0:
            self.SO_even.pop()
            self.SO_even_bymask.pop()
        self.max_rk = max(len(self.Sp), len(self.SO_odd), 2*len(self.SO_even))
